/*  
    Copyright (C) 1999 by Denis Chertykov (denisc@overta.ru)

    You can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

   This program works with this connection scheme.
   
  Connect AVR with parallel port
        AVR                       Parallel Port
   Signal name   Pin             Pin  Signal name
   GND           20 ------------ 19   GND
   GND           20 ------------ 21   GND
   SCK            8 ------------  1   Strobe
   MOSI           6 ------------ 12   Out of paper
   MISO           7 ------------ 11   Busy
   Reset          9 ------------ 16   Init
   Connect parallel port pin 12 with parallel port pin 2 through diod
                                 Out of paper 12 ----|>|--- 2 Data 0
  This connection allow to use send_byte and program AVR flash 
  with Uros Platise `uisp' in `-dapa' mode.

  Bugs: For me this work only if Parellel port in EPP mode.
        May be ECP mode also work ?
        
*/

#include <stdio.h>
#include <io.h>
#include <signal.h>
#include "monitor.h"
#include "mon.h"

#define TIMER0_CK 1
#define EXTREMAL

/* Extract high 8 bit of address */
#define hi8(x) \
 ({ unsigned char t; asm volatile ("ldi %0, hi8 (%1)" :"=r"(t) :"i"(x)); t;})
/* Extract low 8 bit of address */
#define lo8(x) \
 ({ unsigned char t; asm volatile ("ldi %0, lo8 (%1)" :"=r"(t) :"i"(x)); t;})



static struct packet_s packet;
static struct avr_mon_data_s regs;

unsigned char mon_stack[10];
  
static inline unsigned char
get_bit()
{
  unsigned char b;

  /* Waiting for STROBE */
  do
    b = inp (PINB);
  while (b & 0x80);

  /* Set "data readed" */
  cbi (PORTB, 6);

  /* Waiting for STROBE done */
  while (! (inp (PINB) & 0x80));
  
  /* Clear "data readed" */
  sbi (PORTB, 6);
  
  return b & 0x20;
}

static unsigned char
get_byte ()
{
  unsigned char b=0,i;
  for (i=0x80; i; i >>= 1)
    if (get_bit ())
      b |= i;
  return b;
}

int
parse_packet (struct packet_s *pp)
{
  char t;
  t = pp->cmd;
  if (t == 'r')			/* Read memory */
    pp->data = *(unsigned char*)pp->u.addr;
  else if (t == 'w')		/* Write memory */
    *(unsigned char*)pp->u.addr = pp->data;
  else if (t == 'q')		/* Quit (packet.data == 0) */
    ;				/* Or Step (packet.data==1) */
				/* Or Stepping (packet.data == 2) */
  else if (t == 'p')		/* Read program memory */
#ifndef EXTREMAL
    asm volatile ("push r30\n\t"
		  "push r31\n\t"
		  "mov r30, %A1\n\t"
		  "mov r31, %B1\n\t"
		  "lpm\n\t"
		  "mov %0,r0\n\t"
		  "pop r31\n\t"
		  "pop r30"
		  : "=r" (pp->data)
		  : "r" (pp->u.addr));
#else
    asm volatile ("lpm\n\t"
		  "mov %0,r0\n\t"
		  : "=r" (pp->data)
		  : "z" (pp->u.addr));
#endif
  else
    return 'E';
  return pp->cmd;
}

void ret_from_gdb_break ();
void gdb_break ();

/* I don't use SIGNAL(SIG_OVERFLOW0) because it's a special function.
   __attribute__ ((naked)) seems better */

void _overflow0_ (void) __attribute__ ((naked));
void _overflow0_ (void)
{
  asm volatile
    (".global	gdb_break\n"
     "gdb_break:\t"
     "push	r27\n"
     "in	r27, __SREG__\n" /* Save SREG */
     "cli\n"			/* Disable interrupts. Needed for stepping. */
     "sts	regs+28, r28\n"
     "sts	regs+29, r29\n"
     "ldi	r28, lo8 (regs)\n"
     "ldi	r29, hi8 (regs)\n"
     "std	Y+32, r27\n"	/* Put SREG value to his plase */
     "pop	r27\n"		/* Restore previous r27 value */
     "std	Y+27, r27\n"	/* and save it */
     "std	Y+30, r30\n"
     "std	Y+31, r31\n"
     "ldi	r31,0\n"
     "ldi	r30,0\n"
     "push_loop:\n"
     "ld	r27, Z+\n"	/* save registers from 0 to 26 */
     "st	Y+, r27\n"
     "cpi	r30,27\n"
     "brne	push_loop\n"
     "pop	r27\n"		/* pop high byte of ret address hi8(pm(pc)) */
     "pop	r26\n"		/* pop low byte of ret address lo8(pm(pc)) */
     "cpi	r27, hi8 (pm (_overflow0_))\n"
     "ldi	r24, lo8 (pm (_overflow0_))\n"
     "cpc	r26, r24\n"
     "brne	__no_self_trace\n"
     "pop	r27\n"		/* pop high byte of ret address hi8(pm(pc)) */
     "pop	r26\n"		/* pop low byte of ret address lo8(pm(pc)) */
     "__no_self_trace:\t"
     "std	Y+35-27, r26\n"	/* save return address (PC) */
     "std	Y+36-27, r27\n"
     "in	r30, __SP_L__\n"
     "in	r31, __SP_H__\n"
     "std	Y+33-27, r30\n"	/* save SP */
     "std	Y+34-27, r31\n"
     "clr	__zero_reg__\n"
     "ldi	r30, lo8 (mon_stack+8)\n"
     "ldi	r31, hi8 (mon_stack+8)\n"
     "out	__SP_L__, r30\n" /* setup internal monitor stack */
     "out	__SP_H__, r31");

  /* This C text expands to IDEAL assembler equivalent
     I'm happy ;-) */
  if (packet.data == 2)
    {
      char i;
      void *pc = regs.pc;
      void **bp = regs.breakpoint;
    
      for (i = 7; i >= 0; --i)
	if (*bp++ == pc)
	  goto exec_mon;	/* Bad style. but code quality more costly */
      goto next_step;
    }
  exec_mon:
    _monitor_ ();
  next_step:
  asm volatile
    ("ldi	r31,0\n"
     "ldi	r30,0\n"
     "ldi	r28, lo8 (regs)\n"
     "ldi	r29, hi8 (regs)\n"
     "pop_loop:\n"
     "ld	r27, Y+\n"
     "st	Z+, r27\n"
     "cpi	r30, 27\n"
     "brne	pop_loop\n"
     "ldd	r30, Y+6\n"
     "ldd	r31, Y+7\n"
     "out	__SP_L__, r30\n"
     "out	__SP_H__, r31\n"
     "ldd	r27, Y+8\n"
     "push	r27\n"
     "ldd	r27, Y+9\n"
     "push	r27\n"
     "ld	r27, Y\n"
     "ldd	r30, Y+3\n"
     "push	r30\n"		/* real value of r30 */
     "ldd	r30, Y+5\n"	/* r30 = regs.sreg */
     "lds	r31, packet+1\n" /* r31 = packet.data */
     "tst	r31\n"
     "ldd	r31, Y+4\n"
     "ldd	r28, Y+1\n"
     "lds	r29, regs+29\n"
     "brne	_stepping_");
  asm volatile
    ("out	__SREG__, r30\n"
     "pop	r30\n"
     "ret");

  asm volatile ("_stepping_:push r31");
  asm volatile			/* Enable timer interrupts by overflow */
    ("in	r31, %0\n"
     "ori	r31, %1\n"
     "out	%0, r31" :: "I" (TIMSK), "I" (BV (TOIE0)));
  asm volatile			/* set timer 0 counter */
    ("ldi	r31,-4\n"
     "out	%0, r31" :: "I" (TCNT0));
  asm volatile			/* start timer 0 clocks */
    ("ldi	r31, %0\n"
     "out	%1, r31" :: "I"(TIMER0_CK), "I"(TCCR0));
  asm volatile
    ("pop	r31\n"
     "andi	r30,0x7f\n"	/* clear I flag */
     "out	__SREG__, r30\n"
     "pop	r30\n"
     "reti");
}

void _monitor_ (void)
{
  struct packet_s * pp;
  send_byte (0xfc);	 /* Send this byte for switching to monitor mode */
  send_byte (lo8 (&regs));
  send_byte (hi8 (&regs));
  pp = &packet;
  do
    {
      sbi (PORTB, 6);

      /*  outp (0x40, DDRB); */
      sbi (DDRB, 6);		/* We must use sbi and cbi instead of outp */
      cbi (DDRB, 7);
      cbi (DDRB, 5);

      /* Receive command from PC */
      pp->cmd = get_byte ();
      pp->data = get_byte ();
      pp->u.raw.lo = get_byte ();
      pp->u.raw.hi = get_byte ();

      /* Init for sending */
      init_send_byte ();

      pp->cmd = parse_packet (pp);

      send_byte (pp->cmd);	/* Sending back */
      send_byte (pp->data);
      send_byte (pp->u.raw.lo);
      send_byte (pp->u.raw.hi);
    } while (pp->cmd != 'q');

  /* Init for sending */
  init_send_byte ();
}
